"""
Set of commands that helps with integration of Alpa
repository.
"""
from os import getcwd, environ
from pathlib import Path
import re
from subprocess import call
from tempfile import NamedTemporaryFile
from typing import List, Optional
from urllib.parse import urlparse

from click import UsageError, ClickException
import click
from git import Repo, Remote, GitCommandError

from alpa.config.packit import PackitConfig
from alpa.constants import (
    ALPA_FEAT_BRANCH,
    ALPA_FEAT_BRANCH_PREFIX,
    MAIN_BRANCH,
    ORIGIN_NAME,
    UPSTREAM_NAME,
    PackageRequest,
)
from alpa.gh import GithubAPI, GithubRepo
from alpa.messages import (
    CLONED_REPO_IS_NOT_FORK,
    NO_WRITE_ACCESS_ERR,
    NOT_IN_PREDEFINED_STATE,
)


class LocalRepo:
    def __init__(self, repo_path: Path) -> None:
        self.repo_path = repo_path
        # TODO: this will differ in the future with repo_path
        self.repo_root_path = repo_path
        self.local_repo = Repo(str(self.repo_path))
        self.git_cmd = self.local_repo.git

        if not self._is_repo_in_predefined_state():
            raise ClickException(NOT_IN_PREDEFINED_STATE)

    @property
    def remote_associated_with_current_branch(self) -> str:
        return self.local_repo.active_branch.tracking_branch().remote_name

    @property
    def branch(self) -> str:
        return self.local_repo.active_branch.name

    @property
    def package(self) -> str:
        return self.branch.lstrip(ALPA_FEAT_BRANCH_PREFIX)

    @property
    def feat_branch(self) -> str:
        return ALPA_FEAT_BRANCH.format(pkgname=self.package)

    def show_remote_branches(self, remote: str) -> List[str]:
        lines = [
            line.strip()
            for line in self.git_cmd.remote("--verbose", "show", remote).split("\n")
        ]
        # TODO: do a better job
        remote_branch_line = ["Remote branch:", "Remote branches:"]
        start = -1
        for line_to_match in remote_branch_line:
            if line_to_match in lines:
                start = lines.index(line_to_match)
                break

        if start == -1:
            return []

        # TODO: do a better job
        possible_start_of_local_stuff = [
            "Local branch configured for 'git pull':",
            "Local branches configured for 'git pull':",
            "Local ref configured for 'git push':",
            "Local refs configured for 'git push':",
        ]
        end = -1
        for line_to_match in possible_start_of_local_stuff:
            if line_to_match in lines:
                end = lines.index(line_to_match)
                break

        if end == -1:
            end = len(lines)

        lines_with_remote_branches = lines[start + 1 : end]
        return [line.split()[0] for line in lines_with_remote_branches]

    def get_packages(self, regex: str) -> List[str]:
        # self.local_repo.remote(name=UPSTREAM_NAME).refs don't work on every case
        refs_without_main = filter(
            lambda ref: ref != "main", self.show_remote_branches(UPSTREAM_NAME)
        )
        if regex == "":
            return list(refs_without_main)

        pattern = re.compile(regex)
        return [ref for ref in refs_without_main if pattern.match(ref)]

    def _is_repo_in_predefined_state(self) -> bool:
        remotes_name_set = {remote.name for remote in self.local_repo.remotes}
        return remotes_name_set == {ORIGIN_NAME, UPSTREAM_NAME}

    def switch_to_package(self, package: str) -> None:
        try:
            click.echo(self.git_cmd.switch(package))
        except GitCommandError:
            # switching to the package for the first time
            click.echo(f"Switching to the package {package} for the first time")
            click.echo(self.git_cmd.fetch(UPSTREAM_NAME, package))
            click.echo(self.git_cmd.switch(package))

    def get_history_of_branch(self, branch: str, *params: List[str]) -> str:
        return self.git_cmd.log("--decorate", "--graph", *params, branch)

    @staticmethod
    def _get_message_from_editor() -> str:
        with NamedTemporaryFile(suffix=".alpa.tmp") as temp_file:
            call([environ.get("EDITOR", "vim"), temp_file.name])
            temp_file.seek(0)
            output = temp_file.read()
            if isinstance(output, (bytes, bytearray)):
                return output.decode("utf-8")

            return output

    def _ensure_feature_branch(self) -> None:
        if self.branch != self.package:
            return None

        click.echo("Switching to feature branch")
        self.git_cmd.switch("-c", self.feat_branch)

    def commit(self, message: str) -> None:
        self._ensure_feature_branch()
        index = self.local_repo.index
        if message:
            index.commit(message)
        else:
            index.commit(self._get_message_from_editor())

    def add(self, files: List[str]) -> None:
        self._ensure_feature_branch()
        # FIXME: alpa add . acts weird
        self.git_cmd.add(files)

    def pull(self, branch: str) -> None:
        click.echo(self.git_cmd.pull(UPSTREAM_NAME, branch))

    def push(self, branch: str) -> None:
        click.echo(self.git_cmd.push(ORIGIN_NAME, branch))

    def _get_full_reponame(self) -> str:
        for remote in self.local_repo.remotes:
            if remote.name == ORIGIN_NAME:
                return remote.url.split(":")[-1].rstrip(".git")

        return ""

    def create_packit_config(self, override: bool) -> bool:
        packit_conf = PackitConfig(self.package)
        if packit_conf.packit_config_file_exists() and not override:
            return False

        packit_conf.create_packit_config()
        return True


class AlpaRepo(LocalRepo):
    def __init__(self, repo_path: Path, gh_api: Optional[GithubAPI] = None) -> None:
        super().__init__(repo_path)

        self.gh_api = gh_api or GithubAPI()
        namespace, repo_name = self._get_full_reponame().split("/")
        self.gh_repo = self.gh_api.get_repo(namespace, repo_name)

    def create_package(self, package: str) -> None:
        upstream = self.gh_repo.get_upstream()
        if upstream and not upstream.has_write_access(self.gh_api.gh_user):
            raise ClickException(NO_WRITE_ACCESS_ERR)

        self.git_cmd.switch(MAIN_BRANCH)
        self.git_cmd.switch("-c", package)
        self.git_cmd.push(UPSTREAM_NAME, package)
        click.echo(f"Package {package} created")

    def request_package(self, package_name: str) -> None:
        upstream = self.gh_repo.get_root_repo()
        upstream_namespace = upstream.namespace
        issue_repo = self.gh_api.get_repo(upstream_namespace, self.gh_repo.repo_name)
        issue = issue_repo.create_issue(
            PackageRequest.TITLE.value.format(package_name=package_name),
            PackageRequest.BODY.value.format(
                user=self.gh_api.gh_user,
                package_name=package_name,
                repo_name=self.gh_repo.repo_name,
            ),
        )
        issue.add_to_labels(PackageRequest.LABEL)

    def delete_package(self) -> None:
        pass

    @staticmethod
    def _prepare_cloned_repo(local_repo: Repo, gh_repo: GithubRepo) -> None:
        Remote.create(local_repo, UPSTREAM_NAME, gh_repo.upstream_clone_url)

    @staticmethod
    def _get_repo_name_from_url(repo_url: str) -> str:
        return repo_url.split("/")[-1].rstrip(".git")

    @classmethod
    def clone(cls, url: str) -> None:
        # in case of `@` in url -> remove the `git@` prefix form it
        repo_path = urlparse(url.split("@")[-1]).path
        parsed_repo_path = repo_path.strip("/").strip(".git")
        namespace, repo_name = parsed_repo_path.split("/")
        api = GithubAPI()
        gh_repo = api.get_repo(namespace, repo_name)
        if not gh_repo.is_fork():
            raise UsageError(CLONED_REPO_IS_NOT_FORK)

        cloned_repo = Repo.clone_from(
            url, f"{getcwd()}/{cls._get_repo_name_from_url(url)}"
        )
        cls._prepare_cloned_repo(cloned_repo, gh_repo)



"""
File for really long messages.
"""


NO_GH_API_KEY_FOUND = (
    "No GitHub API key found. Please set the "
    "{token} environment variable and pass "
    "the GitHub API token to it. For more information please "
    "see (link TODO)"
)

CLONED_REPO_IS_NOT_FORK = (
    "The url of GitHub repository isn't fork "
    "(does not have upstream). Please provide repository which is "
    "a forked form Alpa repository."
)

NO_WRITE_ACCESS_ERR = (
    "You don't have write access to this repository. If you "
    "want to create new package, please use `alpa request-package "
    "[PACKAGE NAME]` command instead."
)

NOT_IN_PREDEFINED_STATE = (
    "Repository is not in predefined state. Seems like someone "
    "was playing with it. Please see TODO LINK to return this repo "
    "back to its predefined state."
)

RETURNING_CLONE_URL_MSG = (
    "User {user} does not have access to {repo} repository. "
    "Using https url for cloning instead of ssh url."
)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['alpa', 'alpa.cli', 'alpa.config']

package_data = \
{'': ['*']}

install_requires = \
['alpa-conf>=0.3.1',
 'click>=8.0.0',
 'gitpython>=3.0',
 'pygithub>=1.4',
 'pyyaml>=5.0']

entry_points = \
{'console_scripts': ['alpa = alpa.cli.base:entry_point']}

setup_kwargs = {
    'name': 'pyalpa',
    'version': '0.3.1',
    'description': 'Integration tool with Alpa repository',
    'long_description': '## Alpa\n\nAnother cooL way to PAckage in copr\n\n### Status of alpa system\n\n| Package   | Copr build status                                                                                                                                                                                                     |\n| --------- | --------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------- |\n| alpa      | [![Copr build status](https://copr.fedorainfracloud.org/coprs/alpa-team/alpa/package/alpa/status_image/last_build.png)](https://copr.fedorainfracloud.org/coprs/alpa-team/alpa/package/alpa/)                         |\n| alpa-conf | [![Copr build status](https://copr.fedorainfracloud.org/coprs/alpa-team/alpa/package/python-alpa-conf/status_image/last_build.png)](https://copr.fedorainfracloud.org/coprs/alpa-team/alpa/package/python-alpa-conf/) |\n\n## Table of contents\n\n<!-- toc -->\n\n- [Requirements](#requirements)\n- [Installation](#installation)\n- [Usage](#usage)\n- [Contributing](#contributing)\n\n<!-- tocstop -->\n\n### Requirements\n\n- linux OS\n- shell\n- at least python3.8\n- git\n\n### Installation\n\nFrom PyPi:\n\n```bash\n$ pip install --user pyalpa\n```\n\nFrom Copr:\n\n```bash\n$ dnf copr enable alpa-team/alpa\n$ dnf install alpa\n```\n\n### Usage\n\nTODO\n\n### Contributing\n\nTODO\n',
    'author': 'Jiri Kyjovsky',
    'author_email': 'j1.kyjovsky@gmail.com',
    'maintainer': 'Jiří Kyjovský',
    'maintainer_email': 'j1.kyjovsky@gmail.com',
    'url': 'https://github.com/alpa-team/alpa',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from os import getcwd
from pathlib import Path

from yaml import dump

from alpa_conf.constants import PACKIT_CONFIG_NAMES
from alpa_conf.metadata import Metadata


class PackitConfig:
    def __init__(self, package_name: str) -> None:
        self.package_name = package_name
        self.working_dir = Path(getcwd())
        self.metadata = Metadata()

    def get_packit_config(self) -> dict:
        return {
            "specfile_path": f"{self.package_name}.spec",
            "srpm_build_deps": ["pip"],
            "actions": {
                "create-archive": [
                    "pip install pyalpa alpa-conf",
                    'bash -c "alpa get-pkg-archive"',
                    f'bash -c "ls -1 ./{self.package_name}-*.tar.gz"',
                ],
            },
            "jobs": [
                {
                    "job": "copr_build",
                    "trigger": "pull_request",
                    "targets": list(self.metadata.targets),
                },
                {
                    "job": "copr_build",
                    "trigger": "commit",
                    "branch": self.package_name,
                    "targets": list(self.metadata.targets),
                },
            ],
        }

    def packit_config_file_exists(self) -> bool:
        for packit_config_name in PACKIT_CONFIG_NAMES:
            files_in_dir = [
                file.name for file in self.working_dir.iterdir() if file.is_file()
            ]
            if packit_config_name in files_in_dir:
                return True

        return False

    def create_packit_config(self) -> None:
        if self.packit_config_file_exists():
            raise FileExistsError("Packit configuration file already exists")

        with open(".packit.yaml", "w") as packit_yaml:
            packit_yaml.write(dump(self.get_packit_config(), sort_keys=False))
